@@ -58,11 +58,10 @@
 #include "SAPI.h"
 #include "php_apc.h"
 
-#if HAVE_SIGACTION
+#ifdef HAVE_SIGACTION
 #include "apc_signal.h"
 #endif
 
-/* {{{ ZEND_DECLARE_MODULE_GLOBALS(apcu) */
 ZEND_DECLARE_MODULE_GLOBALS(apcu)
 
 /* True globals */
@@ -87,26 +86,10 @@ static void php_apc_init_globals(zend_apcu_globals* apcu_globals)
 	apcu_globals->serializer_name = NULL;
 	apcu_globals->entry_level = 0;
 }
-/* }}} */
 
-/* {{{ PHP_INI */
+/* PHP_INI */
 
-static PHP_INI_MH(OnUpdateShmSegments) /* {{{ */
-{
-	zend_long shm_segments = ZEND_STRTOL(new_value->val, NULL, 10);
-#if APC_MMAP
-	if (shm_segments != 1) {
-		php_error_docref(NULL, E_WARNING, "apc.shm_segments setting ignored in MMAP mode");
-	}
-	APCG(shm_segments) = 1;
-#else
-	APCG(shm_segments) = shm_segments;
-#endif
-	return SUCCESS;
-}
-/* }}} */
-
-static PHP_INI_MH(OnUpdateShmSize) /* {{{ */
+static PHP_INI_MH(OnUpdateShmSize)
 {
 #if PHP_VERSION_ID >= 80200
 	zend_long s = zend_ini_parse_quantity_warn(new_value, entry->name);
@@ -129,18 +112,43 @@ static PHP_INI_MH(OnUpdateShmSize) /* {{{ */
 
 	return SUCCESS;
 }
-/* }}} */
+
+#if defined(APC_MMAP)
+static PHP_INI_MH(OnUpdateMmapHugepageSize)
+{
+	zend_long s;
+
+#if PHP_VERSION_ID >= 80200
+	s = zend_ini_parse_quantity_warn(new_value, entry->name);
+#else
+	s = zend_atol(new_value->val, new_value->len);
+#endif
+
+	if (s < 0) {
+		php_error_docref(NULL, E_CORE_ERROR, "apc.mmap_hugepage_size must be a positive integer");
+		return FAILURE;
+	}
+
+	if (s & (s - 1)) {
+		php_error_docref(NULL, E_CORE_ERROR, "apc.mmap_hugepage_size must be a power of 2");
+		return FAILURE;
+	}
+
+	APCG(mmap_hugepage_size) = s;
+	return SUCCESS;
+}
+#endif
 
 PHP_INI_BEGIN()
 STD_PHP_INI_BOOLEAN("apc.enabled",      "1",    PHP_INI_SYSTEM, OnUpdateBool,              enabled,          zend_apcu_globals, apcu_globals)
-STD_PHP_INI_ENTRY("apc.shm_segments",   "1",    PHP_INI_SYSTEM, OnUpdateShmSegments,       shm_segments,     zend_apcu_globals, apcu_globals)
 STD_PHP_INI_ENTRY("apc.shm_size",       "32M",  PHP_INI_SYSTEM, OnUpdateShmSize,           shm_size,         zend_apcu_globals, apcu_globals)
-STD_PHP_INI_ENTRY("apc.entries_hint",   "4096", PHP_INI_SYSTEM, OnUpdateLong,              entries_hint,     zend_apcu_globals, apcu_globals)
+STD_PHP_INI_ENTRY("apc.entries_hint",   "0",    PHP_INI_SYSTEM, OnUpdateLong,              entries_hint,     zend_apcu_globals, apcu_globals)
 STD_PHP_INI_ENTRY("apc.gc_ttl",         "3600", PHP_INI_SYSTEM, OnUpdateLong,              gc_ttl,           zend_apcu_globals, apcu_globals)
 STD_PHP_INI_ENTRY("apc.ttl",            "0",    PHP_INI_SYSTEM, OnUpdateLong,              ttl,              zend_apcu_globals, apcu_globals)
 STD_PHP_INI_ENTRY("apc.smart",          "0",    PHP_INI_SYSTEM, OnUpdateLong,              smart,            zend_apcu_globals, apcu_globals)
-#if APC_MMAP
-STD_PHP_INI_ENTRY("apc.mmap_file_mask",  NULL,  PHP_INI_SYSTEM, OnUpdateString,            mmap_file_mask,   zend_apcu_globals, apcu_globals)
+#ifdef APC_MMAP
+STD_PHP_INI_ENTRY("apc.mmap_file_mask", NULL,   PHP_INI_SYSTEM, OnUpdateString,            mmap_file_mask,    zend_apcu_globals, apcu_globals)
+STD_PHP_INI_ENTRY("apc.mmap_hugepage_size", "0", PHP_INI_SYSTEM, OnUpdateMmapHugepageSize, mmap_hugepage_size, zend_apcu_globals, apcu_globals)
 #endif
 STD_PHP_INI_BOOLEAN("apc.enable_cli",   "0",    PHP_INI_SYSTEM, OnUpdateBool,              enable_cli,       zend_apcu_globals, apcu_globals)
 STD_PHP_INI_BOOLEAN("apc.slam_defense", "0",    PHP_INI_SYSTEM, OnUpdateBool,              slam_defense,     zend_apcu_globals, apcu_globals)
@@ -150,14 +158,11 @@ STD_PHP_INI_BOOLEAN("apc.use_request_time", "0", PHP_INI_ALL, OnUpdateBool, use_
 STD_PHP_INI_ENTRY("apc.serializer", "php", PHP_INI_SYSTEM, OnUpdateStringUnempty, serializer_name, zend_apcu_globals, apcu_globals)
 PHP_INI_END()
 
-/* }}} */
-
 zend_bool apc_is_enabled(void)
 {
 	return APCG(enabled);
 }
 
-/* {{{ PHP_MINFO_FUNCTION(apcu) */
 static PHP_MINFO_FUNCTION(apcu)
 {
 	php_info_print_table_start();
@@ -168,7 +173,7 @@ static PHP_MINFO_FUNCTION(apcu)
 #else
 	php_info_print_table_row(2, "APCu Debugging", "Disabled");
 #endif
-#if APC_MMAP
+#ifdef APC_MMAP
 	php_info_print_table_row(2, "MMAP Support", "Enabled");
 	php_info_print_table_row(2, "MMAP File Mask", APCG(mmap_file_mask));
 #else
@@ -204,9 +209,7 @@ static PHP_MINFO_FUNCTION(apcu)
 	php_info_print_table_end();
 	DISPLAY_INI_ENTRIES();
 }
-/* }}} */
 
-/* {{{ PHP_MINIT_FUNCTION(apcu) */
 static PHP_MINIT_FUNCTION(apcu)
 {
 #if defined(ZTS) && defined(COMPILE_DL_APCU)
@@ -235,10 +238,12 @@ static PHP_MINIT_FUNCTION(apcu)
 	if (APCG(enabled)) {
 
 		if (!APCG(initialized)) {
-#if APC_MMAP
-			char *mmap_file_mask = APCG(mmap_file_mask);
-#else
 			char *mmap_file_mask = NULL;
+			zend_long mmap_hugepage_size = 0;
+
+#ifdef APC_MMAP
+			mmap_file_mask = APCG(mmap_file_mask);
+			mmap_hugepage_size = APCG(mmap_hugepage_size);
 #endif
 
 			/* ensure this runs only once */
@@ -247,7 +252,7 @@ static PHP_MINIT_FUNCTION(apcu)
 			/* initialize shared memory allocator */
 			apc_sma_init(
 				&apc_sma, (void **) &apc_user_cache, (apc_sma_expunge_f) apc_cache_default_expunge,
-				APCG(shm_segments), APCG(shm_size), mmap_file_mask);
+				APCG(shm_size), APC_ENTRY_SIZE(0), mmap_file_mask, mmap_hugepage_size);
 
 			REGISTER_LONG_CONSTANT(APC_SERIALIZER_CONSTANT, (zend_long)&_apc_register_serializer, CONST_PERSISTENT | CONST_CS);
 
@@ -277,9 +282,7 @@ static PHP_MINIT_FUNCTION(apcu)
 
 	return SUCCESS;
 }
-/* }}} */
 
-/* {{{ PHP_MSHUTDOWN_FUNCTION(apcu) */
 static PHP_MSHUTDOWN_FUNCTION(apcu)
 {
 #define X(str) zend_string_release(apc_str_ ## str);
@@ -300,7 +303,7 @@ static PHP_MSHUTDOWN_FUNCTION(apcu)
 			APCG(initialized) = 0;
 		}
 
-#if HAVE_SIGACTION
+#ifdef HAVE_SIGACTION
 		apc_shutdown_signals();
 #endif
 	}
@@ -309,9 +312,8 @@ static PHP_MSHUTDOWN_FUNCTION(apcu)
 
 	UNREGISTER_INI_ENTRIES();
 	return SUCCESS;
-} /* }}} */
+}
 
-/* {{{ PHP_RINIT_FUNCTION(apcu) */
 static PHP_RINIT_FUNCTION(apcu)
 {
 #if defined(ZTS) && defined(COMPILE_DL_APCU)
@@ -325,15 +327,14 @@ static PHP_RINIT_FUNCTION(apcu)
 			apc_cache_serializer(apc_user_cache, APCG(serializer_name));
 		}
 
-#if HAVE_SIGACTION
+#ifdef HAVE_SIGACTION
 		apc_set_signals();
 #endif
 	}
 	return SUCCESS;
 }
-/* }}} */
 
-/* {{{ proto void apcu_clear_cache() */
+/* proto void apcu_clear_cache() */
 PHP_FUNCTION(apcu_clear_cache)
 {
 	if (zend_parse_parameters_none() == FAILURE) {
@@ -343,9 +344,8 @@ PHP_FUNCTION(apcu_clear_cache)
 	apc_cache_clear(apc_user_cache);
 	RETURN_TRUE;
 }
-/* }}} */
 
-/* {{{ proto array apcu_cache_info([bool limited]) */
+/* proto array apcu_cache_info([bool limited]) */
 PHP_FUNCTION(apcu_cache_info)
 {
 	zend_bool limited = 0;
@@ -360,9 +360,8 @@ PHP_FUNCTION(apcu_cache_info)
 		RETURN_FALSE;
 	}
 }
-/* }}} */
 
-/* {{{ proto array apcu_key_info(string key) */
+/* proto array apcu_key_info(string key) */
 PHP_FUNCTION(apcu_key_info)
 {
 	zend_string *key;
@@ -372,14 +371,11 @@ PHP_FUNCTION(apcu_key_info)
 	ZEND_PARSE_PARAMETERS_END();
 
 	apc_cache_stat(apc_user_cache, key, return_value);
-} /* }}} */
+}
 
-/* {{{ proto array apcu_sma_info([bool limited]) */
+/* proto array apcu_sma_info([bool limited]) */
 PHP_FUNCTION(apcu_sma_info)
 {
-	apc_sma_info_t* info;
-	zval block_lists;
-	int i;
 	zend_bool limited = 0;
 
 	ZEND_PARSE_PARAMETERS_START(0, 1)
@@ -387,15 +383,15 @@ PHP_FUNCTION(apcu_sma_info)
 		Z_PARAM_BOOL(limited)
 	ZEND_PARSE_PARAMETERS_END();
 
-	info = apc_sma_info(&apc_sma, limited);
+	apc_sma_info_t *info = apc_sma_info(&apc_sma, limited);
 
 	if (!info) {
 		php_error_docref(NULL, E_WARNING, "No APC SMA info available.  Perhaps APC is disabled via apc.enabled?");
 		RETURN_FALSE;
 	}
-	array_init(return_value);
 
-	add_assoc_long(return_value, "num_seg", info->num_seg);
+	array_init(return_value);
+	add_assoc_long(return_value, "num_seg", 1);
 	add_assoc_double(return_value, "seg_size", (double)info->seg_size);
 	add_assoc_double(return_value, "avail_mem", (double)apc_sma_get_avail_mem(&apc_sma));
 
@@ -404,30 +400,29 @@ PHP_FUNCTION(apcu_sma_info)
 		return;
 	}
 
-	array_init(&block_lists);
+	/* generate list of free blocks */
+	apc_sma_link_t *p;
+	zval list;
+	array_init(&list);
+	for (p = info->list; p != NULL; p = p->next) {
+		zval link;
 
-	for (i = 0; i < info->num_seg; i++) {
-		apc_sma_link_t* p;
-		zval list;
-
-		array_init(&list);
-		for (p = info->list[i]; p != NULL; p = p->next) {
-			zval link;
-
-			array_init(&link);
-
-			add_assoc_long(&link, "size", p->size);
-			add_assoc_long(&link, "offset", p->offset);
-			add_next_index_zval(&list, &link);
-		}
-		add_next_index_zval(&block_lists, &list);
+		array_init(&link);
+		add_assoc_long(&link, "size", p->size);
+		add_assoc_long(&link, "offset", p->offset);
+		add_next_index_zval(&list, &link);
 	}
+
+	/* Since support for multiple shm segments has been dropped, the "block_lists" array
+	 * only exists to ensure compatibility with existing PHP scripts. */
+	zval block_lists;
+	array_init(&block_lists);
+	add_next_index_zval(&block_lists, &list);
+
 	add_assoc_zval(return_value, "block_lists", &block_lists);
 	apc_sma_free_info(&apc_sma, info);
 }
-/* }}} */
 
-/* {{{ php_apc_update  */
 zend_bool php_apc_update(
 		zend_string *key, apc_cache_atomic_updater_t updater, void *data,
 		zend_bool insert_if_not_found, time_t ttl)
@@ -439,10 +434,7 @@ zend_bool php_apc_update(
 
 	return apc_cache_atomic_update_long(apc_user_cache, key, updater, data, insert_if_not_found, ttl);
 }
-/* }}} */
 
-/* {{{ apc_store_helper(INTERNAL_FUNCTION_PARAMETERS, const zend_bool exclusive)
- */
 static void apc_store_helper(INTERNAL_FUNCTION_PARAMETERS, const zend_bool exclusive)
 {
 	zval *key;
@@ -498,33 +490,24 @@ static void apc_store_helper(INTERNAL_FUNCTION_PARAMETERS, const zend_bool exclu
 		RETURN_FALSE;
 	}
 }
-/* }}} */
 
-/* {{{ proto bool apcu_enabled(void)
-	returns true when apcu is usable in the current environment */
+/* proto bool apcu_enabled(void): returns true when apcu is usable in the current environment */
 PHP_FUNCTION(apcu_enabled) {
 	if (zend_parse_parameters_none() == FAILURE) {
 		return;
 	}
 	RETURN_BOOL(APCG(enabled));
 }
-/* }}} */
 
-/* {{{ proto int apcu_store(mixed key, mixed var [, long ttl ])
- */
+/* proto int apcu_store(mixed key, mixed var [, long ttl ]) */
 PHP_FUNCTION(apcu_store) {
 	apc_store_helper(INTERNAL_FUNCTION_PARAM_PASSTHRU, 0);
 }
-/* }}} */
 
-/* {{{ proto int apcu_add(mixed key, mixed var [, long ttl ])
- */
+/* proto int apcu_add(mixed key, mixed var [, long ttl ]) */
 PHP_FUNCTION(apcu_add) {
 	apc_store_helper(INTERNAL_FUNCTION_PARAM_PASSTHRU, 1);
 }
-/* }}} */
-
-/* {{{ php_inc_updater */
 
 struct php_inc_updater_args {
 	zend_long step;
@@ -537,8 +520,7 @@ static zend_bool php_inc_updater(apc_cache_t *cache, zend_long *entry, void *dat
 	return 1;
 }
 
-/* {{{ proto long apcu_inc(string key [, long step [, bool& success [, long ttl]]])
- */
+/* proto long apcu_inc(string key [, long step [, bool& success [, long ttl]]]) */
 PHP_FUNCTION(apcu_inc) {
 	zend_string *key;
 	struct php_inc_updater_args args;
@@ -567,10 +549,8 @@ PHP_FUNCTION(apcu_inc) {
 
 	RETURN_FALSE;
 }
-/* }}} */
 
-/* {{{ proto long apcu_dec(string key [, long step [, bool &success [, long ttl]]])
- */
+/* proto long apcu_dec(string key [, long step [, bool &success [, long ttl]]]) */
 PHP_FUNCTION(apcu_dec) {
 	zend_string *key;
 	struct php_inc_updater_args args;
@@ -600,19 +580,15 @@ PHP_FUNCTION(apcu_dec) {
 
 	RETURN_FALSE;
 }
-/* }}} */
 
-/* {{{ php_cas_updater */
 static zend_bool php_cas_updater(apc_cache_t *cache, zend_long *entry, void *data) {
 	zend_long *vals = (zend_long *) data;
 	zend_long old = vals[0];
 	zend_long new = vals[1];
 	return ATOMIC_CAS(*entry, old, new);
 }
-/* }}} */
 
-/* {{{ proto int apcu_cas(string key, int old, int new)
- */
+/* proto int apcu_cas(string key, int old, int new) */
 PHP_FUNCTION(apcu_cas) {
 	zend_string *key;
 	zend_long vals[2];
@@ -630,10 +606,8 @@ PHP_FUNCTION(apcu_cas) {
 
 	RETURN_BOOL(apc_cache_atomic_update_long(apc_user_cache, key, php_cas_updater, &vals, 0, 0));
 }
-/* }}} */
 
-/* {{{ proto mixed apcu_fetch(mixed key[, bool &success])
- */
+/* proto mixed apcu_fetch(mixed key[, bool &success]) */
 PHP_FUNCTION(apcu_fetch) {
 	zval *key;
 	zval *success = NULL;
@@ -685,10 +659,8 @@ PHP_FUNCTION(apcu_fetch) {
 		RETURN_FALSE;
 	}
 }
-/* }}} */
 
-/* {{{ proto mixed apcu_exists(mixed key)
- */
+/* proto mixed apcu_exists(mixed key) */
 PHP_FUNCTION(apcu_exists) {
 	zval *key;
 	time_t t;
@@ -728,10 +700,8 @@ PHP_FUNCTION(apcu_exists) {
 		RETURN_FALSE;
 	}
 }
-/* }}} */
 
-/* {{{ proto mixed apcu_delete(mixed keys)
- */
+/* proto mixed apcu_delete(mixed keys) */
 PHP_FUNCTION(apcu_delete) {
 	zval *keys;
 
@@ -780,7 +750,6 @@ PHP_FUNCTION(apcu_entry) {
 
 	apc_cache_entry(apc_user_cache, key, &fci, &fcc, ttl, now, return_value);
 }
-/* }}} */
 
 #ifdef APC_DEBUG
 /* This function is used to test TTL behavior without having to perform sleeps. */
@@ -805,8 +774,7 @@ PHP_FUNCTION(apcu_inc_request_time) {
 }
 #endif
 
-/* {{{ module definition structure */
-
+/* module definition structure */
 zend_module_entry apcu_module_entry = {
 	STANDARD_MODULE_HEADER,
 	PHP_APCU_EXTNAME,
@@ -819,7 +787,6 @@ zend_module_entry apcu_module_entry = {
 	PHP_APCU_VERSION,
 	STANDARD_MODULE_PROPERTIES
 };
-/* }}} */
 
 #ifdef COMPILE_DL_APCU
 ZEND_GET_MODULE(apcu)
@@ -827,7 +794,6 @@ ZEND_GET_MODULE(apcu)
 ZEND_TSRMLS_CACHE_DEFINE();
 #endif
 #endif
-/* }}} */
 
 /*
  * Local variables:
